@@ -1,5 +1,5 @@
 /*
- * Copyright 2005-2006 Luc Verhaegen.
+ * Copyright 2005-2007 Luc Verhaegen.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -34,6 +34,7 @@
 
 #include "via_vgahw.h"
 #include "via_id.h"
+#include "via_output.h"
 #include "via_mode.h"
 
 #include "via_ch7xxx.h"
@@ -183,10 +184,10 @@ static void
 CH7xxxPrintRegs(struct ViaOutput *Output, const char *function)
 {
     CARD8 i, buf;
-    
+
     ViaDebug(Output->scrnIndex, "%s: Printing registers for %s\n",
              function, Output->I2CDev->DevName);
-    
+
     for (i = 0; i < 0x4C; i++) {
 	xf86I2CReadByte(Output->I2CDev, i, &buf);
 	ViaDebug(Output->scrnIndex, "%s %02X: 0x%02X\n",
@@ -213,7 +214,7 @@ CH7011Save(struct ViaOutput *Output)
 
     for (i = 0x1C; i < 0x22; i++)
         xf86I2CReadByte(Output->I2CDev, i, &(Private->Regs[i]));
-    
+
     i = 0x48;
     xf86I2CReadByte(Output->I2CDev, i, &(Private->Regs[i]));
     i = 0x49;
@@ -273,7 +274,7 @@ CH7xxxTVDACSense(struct ViaOutput *Output)
     xf86I2CWriteByte(Output->I2CDev, 0x20, save | 0x01);
 
     Output->I2CDev->pI2CBus->I2CUDelay(Output->I2CDev->pI2CBus, 10);
-    
+
     /* Restore SENSE */
     xf86I2CReadByte(Output->I2CDev, 0x20, &save);
     xf86I2CWriteByte(Output->I2CDev, 0x20, save & 0xFE);
@@ -299,7 +300,7 @@ CH7011DACSense(struct ViaOutput *Output)
 {
     struct CH7xxxOutputPrivate *Private = Output->Private;
     CARD8 sense;
-    
+
     VIAFUNC(Output->scrnIndex);
 
     /* Config already set this */
@@ -340,54 +341,49 @@ CH7011DACSense(struct ViaOutput *Output)
 /*
  * Depends on Output Name, Modes and Private being initialised already.
  */
-static MonPtr
-CH7011Monitor(struct ViaOutput *Output)
+static void
+CH7011TVStandardSet(struct ViaOutput *Output, int Standard)
 {
     struct CH7xxxOutputPrivate *Private = Output->Private;
-    MonPtr Monitor;
 
     VIAFUNC(Output->scrnIndex);
 
-    Monitor = xnfcalloc(1, sizeof(MonRec));
-    memset(Monitor, 0, sizeof(MonRec));
+    if (Output->MonitorName)
+        xfree(Output->MonitorName);
 
-    Monitor->vendor = xnfstrdup("Chrontel");
-    Monitor->model = xnfstrdup(Output->Name);
+    Private->Standard = Standard;
 
-    if (Private->Standard == TVSTANDARD_NTSC) {
-        Monitor->id = xnfstrdup("TV (NTSC)");
+    ViaModesDestroy(Output->Modes);
+    Output->Modes = NULL;
 
-        Monitor->nHsync = 1;
-        Monitor->hsync[0].lo = 31.40;
-        Monitor->hsync[0].hi = 56.70;
+    if (Standard == TVSTANDARD_NTSC) {
+        Output->MonitorName = xnfstrdup("TV (NTSC)");
 
-        Monitor->nVrefresh = 1;
-        Monitor->vrefresh[0].lo = 59.94;
-        Monitor->vrefresh[0].hi = 59.94;
+        Output->numHSync = 1;
+        Output->HSync[0].lo = 31.40;
+        Output->HSync[0].hi = 56.70;
 
-        ViaMonitorAddModetable(Monitor, CH7011TVModesNTSC);
+        Output->numVRefresh = 1;
+        Output->VRefresh[0].lo = 59.94;
+        Output->VRefresh[0].hi = 59.94;
+
+        ViaOutputAddModetable(Output, CH7011TVModesNTSC);
     } else {
-        Monitor->id = xnfstrdup("TV (PAL)");
+        Output->MonitorName = xnfstrdup("TV (PAL)");
 
-        Monitor->nHsync = 1;
-        Monitor->hsync[0].lo = 25.00;
-        Monitor->hsync[0].hi = 68.60;
+        Output->numHSync = 1;
+        Output->HSync[0].lo = 25.00;
+        Output->HSync[0].hi = 68.60;
 
-        Monitor->nVrefresh = 1;
-        Monitor->vrefresh[0].lo = 50.00;
-        Monitor->vrefresh[0].hi = 50.00;
+        Output->numVRefresh = 1;
+        Output->VRefresh[0].lo = 50.00;
+        Output->VRefresh[0].hi = 50.00;
 
-        ViaMonitorAddModetable(Monitor, CH7011TVModesPAL);
+        ViaOutputAddModetable(Output, CH7011TVModesPAL);
     }
 
     Output->ModesExclusive = TRUE;
-
-#ifdef MONREC_HAS_REDUCED
-    /* Make sure that we are not allowing reduced blanking here. */
-    Monitor->reducedblanking = FALSE;
-#endif
-
-    return Monitor;
+    Output->ReducedAllowed = FALSE;
 }
 
 /*
@@ -403,14 +399,14 @@ CH7011TVModeValid(struct ViaOutput *Output, DisplayModePtr mode)
 
     VIAFUNC(Output->scrnIndex);
 
-    if (mode->PrivSize == sizeof(struct CH7011TableRec)) 
+    if (mode->PrivSize == sizeof(struct CH7011TableRec))
         Table = (struct CH7011TableRec *) mode->Private;
-    
+
     if (!Table || strncmp(ID, Table->ID, 12)) {
         xf86DrvMsg(Output->scrnIndex, X_INFO, "Not a mode defined by the TV Encoder.\n");
 	return MODE_BAD;
     }
-    
+
     if (Private->Standard != Table->Standard) {
         if (Private->Standard == TVSTANDARD_NTSC)
             xf86DrvMsg(Output->scrnIndex, X_INFO, "TV standard is NTSC. This is a PAL mode.\n");
@@ -464,11 +460,11 @@ CH7011TVMode(struct ViaOutput *Output, DisplayModePtr mode)
 
     xf86I2CWriteByte(Output->I2CDev, 0x00, Table->Mode);
     xf86I2CWriteByte(Output->I2CDev, 0x01, 0x3F);
-    
+
     /* low pass filter */
     if (Table->Standard == TVSTANDARD_NTSC)
         xf86I2CWriteByte(Output->I2CDev, 0x02, 0x7E); /* default */
-    else 
+    else
         xf86I2CWriteByte(Output->I2CDev, 0x02, 0xE0); /* default */
 
     /* Text Enhancement */
@@ -494,7 +490,7 @@ CH7011TVMode(struct ViaOutput *Output, DisplayModePtr mode)
     /* Black level */
     if (Table->Standard == TVSTANDARD_NTSC)
         xf86I2CWriteByte(Output->I2CDev, 0x07, 0x83); /* default */
-    else 
+    else
         xf86I2CWriteByte(Output->I2CDev, 0x07, 0x6E); /* default */
 
     /* contrast */
@@ -535,7 +531,7 @@ CH7xxxTVPower(struct ViaOutput *Output, Bool On)
     CARD8 save;
 
     VIAFUNC(Output->scrnIndex);
-    
+
     if (On) {
         /* POUTE */
         xf86I2CReadByte(Output->I2CDev, 0x1E, &save);
@@ -545,7 +541,7 @@ CH7xxxTVPower(struct ViaOutput *Output, Bool On)
 	xf86I2CWriteByte(Output->I2CDev, 0x49, 0x20);
     } else {
 	xf86I2CWriteByte(Output->I2CDev, 0x49, 0x3E);
-        
+
         /* !POUTE */
         xf86I2CReadByte(Output->I2CDev, 0x1E, &save);
         xf86I2CWriteByte(Output->I2CDev, 0x1E, save & ~0x02);
@@ -572,7 +568,7 @@ ViaCH7xxxInit(ScrnInfoPtr pScrn, I2CDevPtr pDev)
 
     switch (buf) {
     case 0x17:
-        xf86DrvMsg(pScrn->scrnIndex, X_PROBED, 
+        xf86DrvMsg(pScrn->scrnIndex, X_PROBED,
                    "Detected Chrontel CH7011 TV Encoder\n");
         pDev->DevName = "CH7011";
 
@@ -593,7 +589,7 @@ ViaCH7xxxInit(ScrnInfoPtr pScrn, I2CDevPtr pDev)
         Private->Device = CH7011;
 
         Output->ClockMaster = TRUE;
-        Output->Monitor = CH7011Monitor(Output);
+        CH7011TVStandardSet(Output, Private->Standard);
 
         Output->Save = CH7011Save;
         Output->Restore = CH7011Restore;
