@@ -1,7 +1,5 @@
 /*
- * Copyright 2004-2006 Luc Verhaegen.
- * Copyright 1998-2003 VIA Technologies, Inc. All Rights Reserved.
- * Copyright 2001-2003 S3 Graphics, Inc. All Rights Reserved.
+ * Copyright 2004-2007 Luc Verhaegen.
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
  * copy of this software and associated documentation files (the "Software"),
@@ -41,12 +39,9 @@
 
 #include "via_vgahw.h"
 #include "via_id.h"
+#include "via_output.h"
 #include "via_mode.h"
 
-/* Kill me. */
-#include "via_panel.h"
-
-
 /*
  * Output->Private
  */
@@ -56,7 +51,7 @@ struct ViaPanelOutputPrivate {
 
     int  X;
     int  Y;
-    
+
     /* To enforce Native resolution */
     int  HTotalMax;
     int  HTotalMin;
@@ -70,6 +65,9 @@ struct ViaPanelOutputPrivate {
 
     /* private options */
     Bool  Center;
+
+#define VIA_DI_12BIT  0x00
+#define VIA_DI_24BIT  0x01
     CARD8  BusWidth; /* Digital Output Bus Width */
 };
 
@@ -116,7 +114,9 @@ enum ViaPanelOpts {
 
 static OptionInfoRec ViaPanelOptions[] =
 {
+#ifdef UNUSED
     {OPTION_BUSWIDTH,	 "BusWidth",  	OPTV_ANYSTR,  {0}, FALSE},
+#endif
     {OPTION_CENTER,      "Center",      OPTV_BOOLEAN, {0}, FALSE},
     {OPTION_FORCEPANEL,  "ForcePanel",  OPTV_BOOLEAN, {0}, FALSE}, /* last resort - don't doc */
     {OPTION_PANELSIZE,   "PanelSize",   OPTV_ANYSTR,  {0}, FALSE},
@@ -126,6 +126,16 @@ static OptionInfoRec ViaPanelOptions[] =
 /*
  * Stop the user from telling us lies.
  */
+/* Scratch register contents. */
+#define VIA_PANEL6X4         0
+#define VIA_PANEL8X6         1
+#define VIA_PANEL10X7        2
+#define VIA_PANEL12X7        3
+#define VIA_PANEL12X10       4
+#define VIA_PANEL14X10       5
+#define VIA_PANEL16X12       6
+#define VIA_PANEL_INVALID  255
+
 struct {
     int   X;
     int   Y;
@@ -150,12 +160,13 @@ ViaPanelGetOptions(ScrnInfoPtr pScrn, struct ViaPanelOutputPrivate *Private)
 {
     OptionInfoPtr  Options;
     char  *s = NULL;
-    
+
     Options = xnfalloc(sizeof(ViaPanelOptions));
     memcpy(Options, ViaPanelOptions, sizeof(ViaPanelOptions));
-    
+
     xf86ProcessOptions(pScrn->scrnIndex, pScrn->options, Options);
 
+#ifdef UNUSED
     /* Digital Output Bus Width Option */
     Private->BusWidth = VIA_DI_12BIT;
     if ((s = xf86GetOptValString(Options, OPTION_BUSWIDTH))) {
@@ -169,8 +180,9 @@ ViaPanelGetOptions(ScrnInfoPtr pScrn, struct ViaPanelOutputPrivate *Private)
                        "Digital Output Bus Width is 24BIT\n");
         }
     }
+#endif
 
-    if (VIAPTR(pScrn)->Chipset == VT3108) {
+    if (1) {
         /* primary can't do scaling on its own */
         Private->Center = TRUE;
     } else { /* pretty useless */
@@ -198,7 +210,7 @@ ViaPanelGetOptions(ScrnInfoPtr pScrn, struct ViaPanelOutputPrivate *Private)
         int X, Y;
         if (sscanf(s, "%4dx%4d", &X, &Y) == 2) {
             int i;
-            
+
             for (i = 0; ViaNativePanelSizes[i].X; i++) {
                 if ((X == ViaNativePanelSizes[i].X) &&
                     (Y == ViaNativePanelSizes[i].Y)) {
@@ -225,6 +237,51 @@ ViaPanelGetOptions(ScrnInfoPtr pScrn, struct ViaPanelOutputPrivate *Private)
 }
 
 
+/*
+ * Only run this during PreInit.
+ */
+static Bool
+ViaPanelNativeResolutionFromConsole(ScrnInfoPtr pScrn, CARD16 *HDisplay,
+                                 CARD16 *VDisplay)
+{
+    VIAPtr pVia = VIAPTR(pScrn);
+    vgaHWPtr hwp = VGAHWPTR(pScrn);
+    CARD8 CR6A = hwp->readCrtc(hwp, 0x6A);
+
+    VIAFUNC(pScrn->scrnIndex);
+
+    /* Primary VGA and Panel enabled? */
+    if (xf86IsPrimaryPci(pVia->PciInfo) && (CR6A & 0x08)) {
+        CARD8 CR97 = hwp->readCrtc(hwp, 0x97);
+        CARD8 CR99 = hwp->readCrtc(hwp, 0x99);
+
+        /* Secondary enabled? DFP source is secondary? */
+        if (((CR6A & 0xC0) == 0x40) && (CR97 & 0x10) && (CR99 & 0x10)) {
+            *HDisplay = (hwp->readCrtc(hwp, 0x51) |
+                         ((hwp->readCrtc(hwp, 0x55) & 0x70) << 4)) + 1;
+
+            *VDisplay = (hwp->readCrtc(hwp, 0x59) |
+                         ((hwp->readCrtc(hwp, 0x5D) & 0x38) << 5)) + 1;
+
+            return TRUE;
+        } else if (!(CR97 & 0x10) && !(CR99 & 0x10)) {
+            /* Go for primary timing */
+            *HDisplay = (hwp->readCrtc(hwp, 0x01) + 1) * 8;
+
+            *VDisplay = (hwp->readCrtc(hwp, 0x12) |
+                         ((hwp->readCrtc(hwp, 0x07) & 0x02) << 7) |
+                         ((hwp->readCrtc(hwp, 0x07) & 0x40) << 3) |
+                         ((hwp->readCrtc(hwp, 0x35) & 0x04) << 8)) + 1;
+
+            return TRUE;
+        } else
+            xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "%s: Unhandled combination:"
+                       " CR6A 0x%02X; CR97 0x%02X; CR99 0x%02X\n", __func__,
+                       CR6A, CR97, CR99);
+    }
+    return FALSE;
+}
+
 /*
  *
  */
@@ -238,6 +295,7 @@ ViaPanelNativeResolution(ScrnInfoPtr pScrn,
 
     /* This could have been provided by config */
     if (!Private->X || !Private->Y) {
+        CARD16 HDisplay = 0, VDisplay = 0;
         int i;
 
         for (i = 0; ViaNativePanelSizes[i].X; i++) {
@@ -253,7 +311,24 @@ ViaPanelNativeResolution(ScrnInfoPtr pScrn,
                        "(%d): Contact the driver author ASAP.\n",
                        __func__, pVia->Scratch->PanelSize);
             Private->X = 1024;
-            Private->Y = 768; 
+            Private->Y = 768;
+        }
+
+        if (ViaPanelNativeResolutionFromConsole(pScrn, &HDisplay, &VDisplay)) {
+            if ((Private->X != HDisplay) || (Private->Y != VDisplay)) {
+                if (!pVia->Scratch->PanelSize) {
+                    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "%s: Scratch register"
+                               " panel sizing was probably wrong.\n", __func__);
+                    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "%s: Using resolution"
+                               " of text-mode panel instead.\n", __func__);
+                    Private->X = HDisplay;
+                    Private->Y = VDisplay;
+                } else
+                    xf86DrvMsg(pScrn->scrnIndex, X_INFO, "%s: Scratch based"
+                               "sizing (%dx%d) doesn't match text-mode "
+                               "(%dx%d)\n", __func__, Private->X, Private->Y,
+                               HDisplay, VDisplay);
+            }
         }
     }
 
@@ -261,486 +336,13 @@ ViaPanelNativeResolution(ScrnInfoPtr pScrn,
                Private->X, Private->Y);
 }
 
-
-/*
- * Depends on Output Name, Modes and Private being initialised already.
- */
-static MonPtr
-ViaPanelMonitor(struct ViaOutput *Output, DisplayModePtr Modes)
-{
-    struct ViaPanelOutputPrivate *Private = Output->Private;
-    MonPtr Monitor;
-    int i;
-
-    VIAFUNC(Output->scrnIndex);
-
-    Monitor = xnfcalloc(1, sizeof(MonRec));
-    memset(Monitor, 0, sizeof(MonRec));
-
-    /* Add all modes smaller than native */
-    for (i = 0; Modes[i].name; i++)
-        if ((Private->X >= Modes[i].HDisplay) &&
-            (Private->Y >= Modes[i].VDisplay)) {
-            ViaModesCopyAdd(Monitor, &Modes[i]);
-        }
-
-    if (!Monitor->Modes) {
-        xf86DrvMsg(Output->scrnIndex, X_WARNING,
-                   "%s: No matching mode found.\n", __func__);
-        xfree(Monitor);
-        return NULL;
-    }
-
-    Output->ModesExclusive = TRUE;
-
-    /* set up strings */
-    Monitor->id = xnfstrdup("Panel");
-    Monitor->vendor = xnfstrdup("VIA");
-    Monitor->model = xnfstrdup("TTL Panel");
-
-    /* Set up ranges - Very artificial. */
-    Monitor->nHsync = 1;
-    Monitor->hsync[0].lo = 1024.0;
-    Monitor->hsync[0].hi = 0.0;
-
-    Monitor->nVrefresh = 1;
-    Monitor->vrefresh[0].lo = 1024.0;
-    Monitor->vrefresh[0].hi = 0.0;
-
-    {
-        DisplayModePtr Mode;
-        float tmp;
-        
-        Mode = Monitor->Modes;
-
-        while (Mode) {
-            tmp = ((float) Mode->Clock) / Mode->HTotal;
-
-            if (tmp < Monitor->hsync[0].lo)
-                Monitor->hsync[0].lo = tmp;
-            if (tmp > Monitor->hsync[0].hi)
-                Monitor->hsync[0].hi = tmp;
-
-            tmp = (Mode->Clock * 1000.0) / Mode->HTotal / Mode->VTotal;
-            
-            if (tmp < Monitor->vrefresh[0].lo)
-                Monitor->vrefresh[0].lo = tmp;
-            if (tmp > Monitor->vrefresh[0].hi)
-                Monitor->vrefresh[0].hi = tmp;
-
-            Mode = Mode->next;
-        }
-    }
-
-#ifdef MONREC_HAS_REDUCED
-    /* Our code doesn't care about free modes, but what the heck. */
-    Monitor->reducedblanking = TRUE;
-#endif
-
-    return Monitor;
-}
-
-/*
- *
- * Private->Index is the index to lcdTable.
- *
- */
-static ModeStatus
-ViaPanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
-{
-    struct ViaPanelOutputPrivate *Private = Output->Private;
-    int i;
-    
-    VIAFUNC(Output->scrnIndex);
-
-    if ((mode->PrivSize != sizeof(struct ViaModePriv)) ||
-	(mode->Private != (void *)&ViaPanelPrivate)){
-	xf86DrvMsg(Output->scrnIndex, X_INFO, "%s: Not a mode defined by this "
-                   "drivers panel code.\n", __func__);
-	return MODE_BAD;
-    }
-
-    for (i = 0; i < VIA_PANEL_COUNT; i++)
-	if ((lcdTable[i].X == Private->X) &&
-            (lcdTable[i].Y == Private->Y)) {
-	    int j;
-
-            if ((Private->X == mode->CrtcHDisplay) && 
-                (Private->Y == mode->CrtcVDisplay))
-                return MODE_OK;
-            else if (Private->Center) {
-                for (j = 0; j < VIA_PANEL_COUNT_MAIN; j++)
-                    if (lcdTable[i].CenterMain[j].HDisplay == mode->CrtcHDisplay)
-                        return MODE_OK;
-            } else {
-                for (j = 0; j < VIA_PANEL_COUNT_MAIN; j++) {
-                    if (lcdTable[i].ExpandMain[j].HDisplay == mode->CrtcHDisplay)
-                        return MODE_OK;
-                }
-            }
-
-            xf86DrvMsg(Output->scrnIndex, X_INFO, "%s: Unable to match given "
-                       "mode with this PanelSize.\n", __func__);
-            return MODE_BAD;
-	}
-
-    xf86DrvMsg(Output->scrnIndex, X_INFO, "%s: Unable to match Native"
-               " Resolution with an lcdTable entry.\n", __func__);
-    return MODE_BAD;
-}
-
-
-/*
- *
- */
-static void
-ViaPanelMode(struct ViaOutput *Output, DisplayModePtr mode)
-{
-    ScrnInfoPtr pScrn = xf86Screens[Output->scrnIndex];
-    vgaHWPtr hwp = VGAHWPTR(pScrn);
-    VIAPtr pVia = VIAPTR(pScrn);
-    struct ViaPanelOutputPrivate *Private = Output->Private;
-    VIALCDModeTableRec Table;
-    int  i, HTotal;
-    int PLLPrimary = 0, PLLSecondary = 0;
-    CARD8 Offset;
-
-    VIAFUNC(Output->scrnIndex);
-
-    for (i = 0; i < VIA_PANEL_COUNT; i++)
-	if ((lcdTable[i].X == Private->X) && (lcdTable[i].Y == Private->Y)) {
-            Table = lcdTable[i];
-            break;
-        }
-
-    if (i == VIA_PANEL_COUNT) {
-        xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "%s: unable to find lcdTable"
-                   " index.\n", __func__);
-        return;
-    }
-
-    if ((Private->X == 1280) && (Private->Y == 1024))
-	hwp->writeCrtc(hwp, 0x89, 0x07);
-
-    /* Patch mode for Non-Native. */
-    if ((mode->CrtcHDisplay == Private->X) ||
-        (mode->CrtcVDisplay == Private->Y)) {
-
-        PLLPrimary = PLLSecondary = Table.PLL;
-
-        /* Set up Secondary and Shadow timing. */
-        Offset = 0x50;
-        for (i = 0; i < VIA_PANEL_REG_COUNT_MAIN; i++, Offset++) {
-            if (Offset == 0x60)
-                Offset = 0x65;
-            else if (Offset == 0x68)
-                Offset = 0x6D;
-            hwp->writeCrtc(hwp, Offset, Table.data[i]);
-        }
-
-        hwp->writeCrtc(hwp, 0x77, 0x00);
-        hwp->writeCrtc(hwp, 0x78, 0x00);
-        hwp->writeCrtc(hwp, 0x79, 0x00); 
-
-    } else { /* Patch mode for Non-Native. */
-	struct ViaPanelPatchMain *Patch;
-        int j;
-
-        if (Private->Center)
-	    Patch = Table.CenterMain;
-	else
-            Patch = Table.ExpandMain;
-
-	/* Set LCD Mode patch registers. */
-	for (i = 0; i < VIA_PANEL_COUNT_MAIN; i++, Patch++)
-	    if ((Patch->HDisplay == mode->CrtcHDisplay) &&
-                (Patch->VDisplay == mode->CrtcVDisplay))
-                break;
-        
-        if (i == VIA_PANEL_COUNT_MAIN) {
-            xf86DrvMsg(pScrn->scrnIndex, X_ERROR, "%s: unable to find matching"
-                       " secondary Patch index.\n", __func__);
-            return;
-        }
-            
-               
-        PLLPrimary = Patch->PLLPrimary;
-        PLLSecondary = Patch->PLLSecondary;
-
-        /* Set up Secondary and Shadow timing. */
-        Offset = 0x50;
-        for (i = 0; i < VIA_PANEL_REG_COUNT_MAIN; i++, Offset++) {
-            if (Offset == 0x60)
-                Offset = 0x65;
-            else if (Offset == 0x68)
-                Offset = 0x6D;
-            hwp->writeCrtc(hwp, Offset, Patch->data[i]);
-        }
-
-        if (Patch->Scaled) {
-            int Scale;
-            
-            if (mode->CrtcHDisplay != Private->X) {
-                Scale = ((float) ((mode->CrtcHDisplay - 1) * 1024) /
-                         ((float) Private->X)) + 0.5;
-                
-                hwp->writeCrtc(hwp, 0x77, Scale & 0xFF);
-                ViaCrtcMask(hwp, 0x79, (Scale & 0x300) >> 4, 0x30);
-                
-                ViaCrtcMask(hwp, 0x79, 0x02, 0x02);
-            } else {
-                hwp->writeCrtc(hwp, 0x77, 0x00);
-                ViaCrtcMask(hwp, 0x79, 0x00, 0x30);
-            }
-            
-            if (mode->CrtcVDisplay != Private->Y) {
-                Scale = ((float) ((mode->CrtcVDisplay - 1) * 1024) /
-                         ((float) Private->Y)) + 0.5;
-                
-                hwp->writeCrtc(hwp, 0x78, Scale & 0xFF);
-                ViaCrtcMask(hwp, 0x79, (Scale & 0x300) >> 2, 0xC0);
-                
-                ViaCrtcMask(hwp, 0x79, 0x04, 0x04);
-            } else {
-                hwp->writeCrtc(hwp, 0x78, 0x00);
-                ViaCrtcMask(hwp, 0x79, 0x00, 0xC0);
-            }
-            
-            ViaCrtcMask(hwp, 0x79, 0x01, 0x01);
-        } else {
-            hwp->writeCrtc(hwp, 0x77, 0x00);
-            hwp->writeCrtc(hwp, 0x78, 0x00);
-            hwp->writeCrtc(hwp, 0x79, 0x00);
-        }
-
-	/* Set LCD Mode Patch registers when Primary. */
-	if (!pVia->IsSecondary) {
-            struct ViaPanelPatchPrimary *Primary;
-
-            if (Private->Center)
-                Primary = Table.CenterPrimary;
-            else
-                Primary = Table.ExpandPrimary;
-
-	    for (i = 0; i < VIA_PANEL_COUNT_PRIMARY; i++, Primary++) {
-                if ((Primary->HDisplay == mode->CrtcHDisplay) &&
-                    (Primary->VDisplay == mode->CrtcVDisplay)) {
-		    for (j = 0; j < Primary->numEntry; j++)
-                        hwp->writeCrtc(hwp, Primary->offset[j], Primary->data[j]);
-		    break;
-		}
-	    }
-	}
-    }
-
-    /* LCD patch 3D5.02 */
-    HTotal = hwp->readCrtc(hwp, 0x01);
-    hwp->writeCrtc(hwp, 0x02, HTotal);
-
-    /* Set up PLL */
-    {
-        CARD32 PLL;
-
-        if (Private->BusWidth == VIA_DI_12BIT) {
-            PLLPrimary *= 2;
-            PLLSecondary *= 2;
-        }
-
-        PLL = ViaPLLGenerate(pScrn, PLLSecondary);
-        ViaSetSecondaryDotclock(pScrn, PLL);
-
-        if (!pVia->IsSecondary) {
-            if (PLLPrimary == PLLSecondary)
-                pVia->PanelClock = PLL;
-            else
-                pVia->PanelClock = ViaPLLGenerate(pScrn, PLLPrimary);
-        }
-    }
-
-    /* Enable LCD */
-    if (!pVia->IsSecondary) {
-        /* Enable Simultaneous */
-        if (Private->BusWidth == VIA_DI_12BIT) {
-	    hwp->writeCrtc(hwp, 0x6B, 0xA8);
-
-            if ((pVia->Chipset == VT3122) && VT3122_REV_IS_AX(pVia->ChipRev))
-		hwp->writeCrtc(hwp, 0x93, 0xB1);
-            else
-		hwp->writeCrtc(hwp, 0x93, 0xAF);
-        } else {
-	    ViaCrtcMask(hwp, 0x6B, 0x08, 0x08);
-	    hwp->writeCrtc(hwp, 0x93, 0x00);
-        }
-	hwp->writeCrtc(hwp, 0x6A, 0x48);
-    } else {
-        /* Enable SAMM */
-        if (Private->BusWidth == VIA_DI_12BIT) {
-	    ViaCrtcMask(hwp, 0x6B, 0x20, 0x20);
-            if ((pVia->Chipset == VT3122) && VT3122_REV_IS_AX(pVia->ChipRev))
-		hwp->writeCrtc(hwp, 0x93, 0xB1);
-            else
-		hwp->writeCrtc(hwp, 0x93, 0xAF);
-        } else {
-	    hwp->writeCrtc(hwp, 0x6B, 0x00);
-	    hwp->writeCrtc(hwp, 0x93, 0x00);
-        }
-	hwp->writeCrtc(hwp, 0x6A, 0xC8);
-    }
-
-    Output->Power(Output, TRUE);
-}
-
-
-/*
- *
- */
-static void
-ViaPanelPower(struct ViaOutput *Output, Bool On)
-{
-    ScrnInfoPtr pScrn = xf86Screens[Output->scrnIndex];
-    vgaHWPtr hwp = VGAHWPTR(pScrn);
-    VIAPtr pVia = VIAPTR(pScrn);
-
-    ViaDebug(Output->scrnIndex, "%s: %s.\n", __func__, On ? "On" : "Off");
-
-    /* Enable LCD */
-    if (On)
-	ViaCrtcMask(hwp, 0x6A, 0x08, 0x08);
-    else
-	ViaCrtcMask(hwp, 0x6A, 0x00, 0x08);
-
-    usleep(1);
-
-    if (pVia->Chipset == VT3122) {
-        if (On) {
-            ViaCrtcMask(hwp, 0x91, 0x10, 0x10);
-            usleep(0x19);
-            ViaCrtcMask(hwp, 0x91, 0x08, 0x08);
-            usleep(0x1FE);
-            ViaCrtcMask(hwp, 0x91, 0x06, 0x06);
-            usleep(0x01);
-        } else {
-            ViaCrtcMask(hwp, 0x91, 0x00, 0x06);
-            usleep(0xD2);
-            ViaCrtcMask(hwp, 0x91, 0x00, 0x08);
-            usleep(0x19);
-            ViaCrtcMask(hwp, 0x91, 0x00, 0x10);
-            usleep(0x01);
-        }
-#if 0
-        if (!Private->X || !Private->Y) {  /* old index 0x00 -- never used */
-            if (On) {
-                ViaCrtcMask(hwp, 0x91, 0x10, 0x10);
-                usleep(0x19);
-                ViaCrtcMask(hwp, 0x91, 0x08, 0x08);
-                usleep(0x64);
-            } else {
-                ViaCrtcMask(hwp, 0x91, 0x00, 0x18);
-                usleep(0x01);
-            }
-        }
-#endif
-    } else { /* VT7205 and up */
-        if (On) {
-            ViaCrtcMask(hwp, 0x91, 0x10, 0x10);
-            usleep(0x19);
-            ViaCrtcMask(hwp, 0x91, 0x08, 0x08);
-            usleep(0x1FE);
-            ViaSeqMask(hwp, 0x3D, 0x20, 0x20);
-            usleep(0x01);
-        } else {
-            ViaSeqMask(hwp, 0x3D, 0x00, 0x20);
-            usleep(0xD2);
-            ViaCrtcMask(hwp, 0x91, 0x00, 0x08);
-            usleep(0x19);
-            ViaCrtcMask(hwp, 0x91, 0x00, 0x10);
-            usleep(0x01);
-        }
-    }
-
-    usleep(1);
-}
-
-/*
- *
- */
-static struct ViaOutput *
-VT3122PanelInit(ScrnInfoPtr pScrn, struct ViaOutput *Output)
-{
-    Output->Name = "Panel";
-
-    Output->ClockMaster = FALSE;
-
-    Output->Monitor = ViaPanelMonitor(Output, ViaPanelModes);
-
-    Output->Save = NULL;
-    Output->Restore = NULL;
-    Output->Sense = NULL;
-    Output->ModeValid = ViaPanelModeValid;
-    Output->Mode = ViaPanelMode;
-    Output->Power = ViaPanelPower;
-    Output->PrintRegs = NULL;
-
-    return Output;
-}
-
-/*
- *
- * Primary only Panel implementation for the VT3108.
- *
- */
-static struct {
-    int X;
-    int Y;
-    char Name[10];
-} VT3108PanelModes[] = {
-    {1280, 768, "1280x768"},
-    {1024, 768, "1024x768"},
-    { 800, 600, "800x600"},
-    { 640, 480, "640x480"},
-    {   0,   0, ""},
-};
-
-/*
- * Add centered modes.
- *
- * We use Mode destructively here.
- *
- */
-static void
-VT3108PanelMonitorCenteredAdd(struct ViaOutput *Output, MonPtr Monitor,
-                              DisplayModePtr Mode)
-{
-    struct ViaPanelOutputPrivate *Private = Output->Private;
-    int i;
-
-    for (i = 0; VT3108PanelModes[i].X; i++) {
-        if ((VT3108PanelModes[i].X > Private->X) ||
-            (VT3108PanelModes[i].Y > Private->Y))
-            continue;
-
-        /* We already added native resolution */
-        if ((VT3108PanelModes[i].X == Private->X) &&
-            (VT3108PanelModes[i].Y == Private->Y))
-            continue;
-
-        Mode->HDisplay = VT3108PanelModes[i].X;
-        Mode->VDisplay = VT3108PanelModes[i].Y;
-        xfree(Mode->name);
-        Mode->name = xnfstrdup(VT3108PanelModes[i].Name);
-        
-        ViaPrintModeline(Output->scrnIndex, Mode);
-        ViaModesCopyAdd(Monitor, Mode);
-    }
-}
-
 /*
  * Here we assume that panels are capable of at least 57.5 and 65Hz CVT timing
  * and CVT reduced blanking timing. We use those timings to impose our limits.
  *
  */
 static void
-VT3108PanelMonitorCenteredTiming(struct ViaOutput *Output, MonPtr Monitor)
+ViaPanelCenteredTiming(struct ViaOutput *Output)
 {
     struct ViaPanelOutputPrivate *Private = Output->Private;
     DisplayModePtr Mode, RMode;
@@ -749,7 +351,7 @@ VT3108PanelMonitorCenteredTiming(struct ViaOutput *Output, MonPtr Monitor)
     Mode = ViaCVTMode(Private->X, Private->Y, 65.0, FALSE, FALSE);
 
     /* Reduced blanking mode = lower limit for timing */
-    RMode = ViaCVTMode(Private->X, Private->Y, 60.0, TRUE, FALSE); 
+    RMode = ViaCVTMode(Private->X, Private->Y, 60.0, TRUE, FALSE);
 
     Private->HTotalMax = Mode->HTotal;
     Private->HTotalMin = RMode->HTotal;
@@ -764,76 +366,117 @@ VT3108PanelMonitorCenteredTiming(struct ViaOutput *Output, MonPtr Monitor)
     ViaModesDestroy(RMode);
 
     /* Safe bets */
-    Monitor->nVrefresh = 1;
-    Monitor->vrefresh[0].lo = 57.5;
-    Monitor->vrefresh[0].hi = 65.0;
+    Output->numVRefresh = 1;
+    Output->VRefresh[0].lo = 57.5;
+    Output->VRefresh[0].hi = 65.0;
 
-    /* now get slow timing; for hsync */
+    /* now get slow timing; for HSync */
     RMode = ViaCVTMode(Private->X, Private->Y, 57.5, FALSE, FALSE);
 
-    Monitor->nHsync = 1;
-    Monitor->hsync[0].hi = (float) Mode->Clock / Mode->HTotal;
-    Monitor->hsync[0].lo = (float) RMode->Clock / RMode->HTotal;
+    Output->numHSync = 1;
+    Output->HSync[0].hi = (float) Mode->Clock / Mode->HTotal;
+    Output->HSync[0].lo = (float) RMode->Clock / RMode->HTotal;
 
     ViaModesDestroy(Mode);
     ViaModesDestroy(RMode);
 }
 
 /*
+ *
+ * Primary only Panel implementation.
  *
  */
-static MonPtr
-VT3108PanelMonitor(struct ViaOutput *Output, DisplayModePtr Modes)
+static struct {
+    int X;
+    int Y;
+    char Name[10];
+} ViaPanelModes[] = {
+    {1280, 768, "1280x768"},
+    {1024, 768, "1024x768"},
+    { 800, 600, "800x600"},
+    { 640, 480, "640x480"},
+    {   0,   0, ""},
+};
+
+/*
+ * Add centered modes.
+ *
+ * We use Mode destructively here.
+ *
+ */
+static void
+ViaPanelCenteredAdd(struct ViaOutput *Output,  DisplayModePtr Mode)
+{
+    struct ViaPanelOutputPrivate *Private = Output->Private;
+    int i;
+
+    for (i = 0; ViaPanelModes[i].X; i++) {
+        if ((ViaPanelModes[i].X > Private->X) ||
+            (ViaPanelModes[i].Y > Private->Y))
+            continue;
+
+        /* We already added native resolution */
+        if ((ViaPanelModes[i].X == Private->X) &&
+            (ViaPanelModes[i].Y == Private->Y))
+            continue;
+
+        Mode->HDisplay = ViaPanelModes[i].X;
+        Mode->VDisplay = ViaPanelModes[i].Y;
+        xfree(Mode->name);
+        Mode->name = xnfstrdup(ViaPanelModes[i].Name);
+
+        Mode->type = M_T_DRIVER;
+
+        ViaOutputModesCopyAdd(Output, Mode);
+    }
+}
+
+/*
+ *
+ */
+static void
+ViaPanelTimingSet(struct ViaOutput *Output)
 {
     struct ViaPanelOutputPrivate *Private = Output->Private;
-    MonPtr Monitor;
     DisplayModePtr Mode;
 
     VIAFUNC(Output->scrnIndex);
 
-    Monitor = xnfcalloc(1, sizeof(MonRec));
-    memset(Monitor, 0, sizeof(MonRec));
-
-#ifdef MONREC_HAS_REDUCED
     /* We need to add both CVT as normal modes when we can. */
-    Monitor->reducedblanking = TRUE;
+    Output->ReducedAllowed = TRUE;
 
     /* Native mode */
     Mode = ViaCVTMode(Private->X, Private->Y, 60.0, TRUE, FALSE);
-    ViaModesCopyAdd(Monitor, Mode);
+    Mode->type = M_T_DRIVER | M_T_PREFERRED;
+    ViaOutputModesCopyAdd(Output, Mode);
 
     /* Add centered Modes */
-    VT3108PanelMonitorCenteredAdd(Output, Monitor, Mode);
+    ViaPanelCenteredAdd(Output, Mode);
 
     ViaModesDestroy(Mode);
-#endif
 
     /* Native mode */
     Mode = ViaCVTMode(Private->X, Private->Y, 60.0, FALSE, FALSE);
-    ViaModesCopyAdd(Monitor, Mode);
+    Mode->type = M_T_DRIVER | M_T_PREFERRED;
+    ViaOutputModesCopyAdd(Output, Mode);
 
     /* Add centered Modes */
-    VT3108PanelMonitorCenteredAdd(Output, Monitor, Mode);
+    ViaPanelCenteredAdd(Output, Mode);
 
     ViaModesDestroy(Mode);
 
-    /* set up strings */
-    Monitor->id = xnfstrdup("Panel");
-    Monitor->vendor = xnfstrdup("VIA");
-    Monitor->model = xnfstrdup("TTL Panel");
+    Output->MonitorName = xnfstrdup("TTL Panel");
 
-    VT3108PanelMonitorCenteredTiming(Output, Monitor);
+    ViaPanelCenteredTiming(Output);
 
     Output->ModesExclusive = FALSE;
-
-    return Monitor;
 }
 
 /*
  *
  */
 static ModeStatus
-VT3108PanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
+ViaPanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
 {
     struct ViaPanelOutputPrivate *Private = Output->Private;
     int Sync;
@@ -843,11 +486,11 @@ VT3108PanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
     if ((mode->HDisplay > Private->X) || (mode->VDisplay > Private->Y))
         return MODE_PANEL;
 
-    if ((mode->CrtcHTotal > Private->HTotalMax) || 
+    if ((mode->CrtcHTotal > Private->HTotalMax) ||
         (mode->CrtcHTotal < Private->HTotalMin))
         return MODE_H_ILLEGAL;
 
-    if ((mode->CrtcVTotal > Private->VTotalMax) || 
+    if ((mode->CrtcVTotal > Private->VTotalMax) ||
         (mode->CrtcVTotal < Private->VTotalMin))
         return MODE_V_ILLEGAL;
 
@@ -863,6 +506,9 @@ VT3108PanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
     if (Sync < Private->VSyncMin)
         return MODE_VSYNC_NARROW;
 
+    if ((mode->HDisplay == Private->X) || (mode->VDisplay == Private->Y))
+        return MODE_OK;
+
     if (Private->Center) {
         int Shift;
 
@@ -875,21 +521,31 @@ VT3108PanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
          * Sync is shifted too so that an attached CRT will be centered as well.
          */
 
-        Shift = (Private->X - mode->HDisplay) / 2;
-        if (Shift) {
+        Shift = ((Private->X - mode->HDisplay) / 2) & ~0x07;
+        if (Shift && ((mode->CrtcHBlankEnd != (mode->CrtcHTotal - Shift)) ||
+                      (mode->CrtcHBlankStart != mode->CrtcHDisplay + Shift))) {
             mode->CrtcHBlankEnd = mode->CrtcHTotal - Shift;
             mode->CrtcHBlankStart = mode->CrtcHDisplay + Shift;
-            mode->CrtcHSyncStart -= Shift / 2;
-            mode->CrtcHSyncEnd -= Shift / 2;
+
+            /* calculate new sync posisition */
+            Shift = mode->CrtcHSyncEnd - mode->CrtcHSyncStart;
+            mode->CrtcHSyncEnd = ((mode->CrtcHTotal + mode->CrtcHDisplay) & ~0x0F) >> 1;
+            mode->CrtcHSyncStart = mode->CrtcHSyncEnd - Shift;
+
             mode->CrtcHAdjusted = TRUE;
         }
 
         Shift = (Private->Y - mode->VDisplay) / 2;
-        if (Shift) {
+        if (Shift && ((mode->CrtcVBlankEnd != (mode->CrtcVTotal - Shift)) ||
+                      (mode->CrtcVBlankStart != mode->CrtcVDisplay + Shift))) {
             mode->CrtcVBlankEnd = mode->CrtcVTotal - Shift;
             mode->CrtcVBlankStart = mode->CrtcVDisplay + Shift;
-            mode->CrtcVSyncStart -= Shift / 2;
-            mode->CrtcVSyncEnd -= Shift / 2;
+
+             /* calculate new sync posisition */
+            Shift = mode->CrtcVSyncEnd - mode->CrtcVSyncStart;
+            mode->CrtcVSyncEnd = (mode->CrtcVTotal + mode->CrtcVDisplay) >> 1;
+            mode->CrtcVSyncStart = mode->CrtcVSyncEnd - Shift;
+
             mode->CrtcVAdjusted = TRUE;
         }
     }
@@ -901,10 +557,10 @@ VT3108PanelModeValid(struct ViaOutput *Output, DisplayModePtr mode)
  *
  */
 static void
-VT3108PanelMode(struct ViaOutput *Output, DisplayModePtr mode)
+ViaPanelMode(struct ViaOutput *Output, DisplayModePtr mode)
 {
     ScrnInfoPtr pScrn = xf86Screens[Output->scrnIndex];
-    vgaHWPtr hwp = VGAHWPTR(pScrn); 
+    vgaHWPtr hwp = VGAHWPTR(pScrn);
     struct ViaPanelOutputPrivate *Private = Output->Private;
 
     VIAFUNC(Output->scrnIndex);
@@ -930,6 +586,77 @@ VT3108PanelMode(struct ViaOutput *Output, DisplayModePtr mode)
     }
 }
 
+/*
+ * Reduction of old VIA code, untested.
+ */
+static void
+VT3122PanelPower(struct ViaOutput *Output, Bool On)
+{
+    vgaHWPtr hwp = VGAHWPTR(xf86Screens[Output->scrnIndex]);
+
+    ViaDebug(Output->scrnIndex, "%s: %s.\n", __func__, On ? "On" : "Off");
+
+    /* Enable LCD */
+    if (On)
+	ViaCrtcMask(hwp, 0x6A, 0x08, 0x08);
+    else
+	ViaCrtcMask(hwp, 0x6A, 0x00, 0x08);
+
+    usleep(1);
+
+    if (On) {
+        ViaCrtcMask(hwp, 0x91, 0x10, 0x10);
+        usleep(0x19);
+        ViaCrtcMask(hwp, 0x91, 0x08, 0x08);
+        usleep(0x1FE);
+        ViaCrtcMask(hwp, 0x91, 0x06, 0x06);
+        usleep(0x01);
+    } else {
+        ViaCrtcMask(hwp, 0x91, 0x00, 0x06);
+        usleep(0xD2);
+        ViaCrtcMask(hwp, 0x91, 0x00, 0x08);
+        usleep(0x19);
+        ViaCrtcMask(hwp, 0x91, 0x00, 0x10);
+        usleep(0x01);
+    }
+}
+
+/*
+ * Reduction of old VIA code, untested.
+ */
+static void
+VT7205PanelPower(struct ViaOutput *Output, Bool On)
+{
+    vgaHWPtr hwp = VGAHWPTR(xf86Screens[Output->scrnIndex]);
+
+    ViaDebug(Output->scrnIndex, "%s: %s.\n", __func__, On ? "On" : "Off");
+
+    /* Enable LCD */
+    if (On)
+	ViaCrtcMask(hwp, 0x6A, 0x08, 0x08);
+    else
+	ViaCrtcMask(hwp, 0x6A, 0x00, 0x08);
+
+    usleep(1);
+
+    if (On) {
+        ViaCrtcMask(hwp, 0x91, 0x10, 0x10);
+        usleep(0x19);
+        ViaCrtcMask(hwp, 0x91, 0x08, 0x08);
+        usleep(0x1FE);
+        ViaSeqMask(hwp, 0x3D, 0x20, 0x20);
+        usleep(0x01);
+    } else {
+        ViaSeqMask(hwp, 0x3D, 0x00, 0x20);
+        usleep(0xD2);
+        ViaCrtcMask(hwp, 0x91, 0x00, 0x08);
+        usleep(0x19);
+        ViaCrtcMask(hwp, 0x91, 0x00, 0x10);
+        usleep(0x01);
+    }
+
+    usleep(1);
+}
 
 /*
  * This implements only superficial power down. Good enough for DPMS.
@@ -944,37 +671,13 @@ VT3108PanelPower(struct ViaOutput *Output, Bool On)
 
     ViaCrtcMask(hwp, 0x91, 0x00, 0x01); /* Disable Software Control */
     usleep(1);
-    
+
     if (On) /* Direct Panel Display and Backlight control */
         ViaCrtcMask(hwp, 0x91, 0x00, 0xC0);
     else
         ViaCrtcMask(hwp, 0x91, 0xC0, 0xC0);
 }
 
-/*
- *
- */
-static struct ViaOutput *
-VT3108PanelInit(ScrnInfoPtr pScrn, struct ViaOutput *Output)
-{
-    Output->Name = "Panel";
-
-    Output->ClockMaster = FALSE;
-
-    Output->Monitor = VT3108PanelMonitor(Output, ViaPanelModes);
-
-    Output->Save = NULL;
-    Output->Restore = NULL;
-    Output->Sense = NULL;
-    Output->ModeValid = VT3108PanelModeValid;
-    Output->Mode = VT3108PanelMode;
-    Output->Power = VT3108PanelPower;
-    Output->PrintRegs = NULL;
-
-    return Output;
-}
-
-
 /*
  * future: I2CDevPtr for LVDS encoders.
  */
@@ -990,14 +693,14 @@ ViaPanelInit(ScrnInfoPtr pScrn, I2CDevPtr pDev)
     if (pDev)
         xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "%s: Ignoring I2C Device"
                    " passed.\n", __func__);
-    
+
     Output = xnfcalloc(1, sizeof(struct ViaOutput));
 
     Output->Prev = NULL;
     Output->Next = NULL;
     Output->scrnIndex = pScrn->scrnIndex;
     Output->I2CDev = pDev;
-    Output->Type = OUTPUT_PANEL; 
+    Output->Type = OUTPUT_PANEL;
 
     Private = ViaPanelPrivateCreate(Output);
 
@@ -1012,7 +715,7 @@ ViaPanelInit(ScrnInfoPtr pScrn, I2CDevPtr pDev)
             Private->Present = TRUE;
         } else {
             Output->PrivateDestroy(Output);
-            
+
             xfree(Output->Options);
             xfree(Output);
             return NULL;
@@ -1021,12 +724,29 @@ ViaPanelInit(ScrnInfoPtr pScrn, I2CDevPtr pDev)
 
     ViaPanelNativeResolution(pScrn, Private);
 
+    Output->Name = "Panel";
+
+    Output->ClockMaster = FALSE;
+
+    ViaPanelTimingSet(Output);
+
+    Output->Save = NULL;
+    Output->Restore = NULL;
+    Output->Sense = NULL;
+    Output->ModeValid = ViaPanelModeValid;
+    Output->Mode = ViaPanelMode;
+    Output->PrintRegs = NULL;
+
     switch (pVia->Chipset) {
     case VT3122:
+        Output->Power = VT3122PanelPower;
+        break;
     case VT7205:
-        return VT3122PanelInit(pScrn, Output);
+        Output->Power = VT7205PanelPower;
+        break;
     case VT3108:
-        return VT3108PanelInit(pScrn, Output);
+        Output->Power = VT3108PanelPower;
+        break;
     default:
 	xf86DrvMsg(pScrn->scrnIndex, X_WARNING, "Panel on %s is currently "
                    "not supported.\n", pScrn->chipset);
@@ -1035,4 +755,6 @@ ViaPanelInit(ScrnInfoPtr pScrn, I2CDevPtr pDev)
         xfree(Output);
         return NULL;
     }
+
+    return Output;
 }
